#============================================================
#          COLUMNAS MOVEDIZAS
#
#  Manuel Collado, http://mcollado.z15.es
#  Marzo, 2020 - Public domain
#
#============================================================
package require Tk
set rootdir [file normalize [file dirname [info script]]/..]
source $rootdir/lib/common.tcl

#============================================================
#          Modelo de datos
#============================================================

#      Estado del juego
# -----------------------------------------------------------
set fichero_texto "$rootdir/textos/texto.txt"
set numfilas 8
set numcolumnas 27
set longtexto [expr $numfilas * $numcolumnas]
set texto ""            ;# texto a recomponer
set posicion0 [list]    ;# posicin barajada de cada columna
set posicion1 [list]    ;# posiciones dadas por el jugador
set posicion2 [list]    ;# inversa de 'posicin1'
set seleccion_alto -1   ;# columna superior seleccionada
set seleccion_bajo -1   ;# columna original seleccionada


# Elegir fichero con textos
#
proc nuevo_fichero {} {
    global fichero_texto
    set fichero [new_file $fichero_texto]
    if {$fichero != ""} {
        set fichero_texto $fichero
       .texto.fichero.nombre configure -text [file tail $fichero_texto]
        nuevo_texto
    }
}

# mostrar texto por columnas en el orden indicado    *********** PRUEBA ************
proc mostrar_texto {texto columnas} {
    set numcols [llength $columnas]
    set numchars [string length $texto]
    set numfilas [expr int(($numchars+$numcols-1)/$numcols)]
    puts ""
    puts $columnas
    for {set fil 0} {$fil<$numfilas} {incr fil} {
        for {set col 0} {$col<$numcols} {incr col} {
            set k [lindex $columnas $col]
            if {$k>=0} {
            set off [expr $fil*$numcols + [lindex $columnas $col]]
                puts -nonewline " [string index $texto $off]"
            } else {
                puts -nonewline "  "
            }
        }
        puts ""
    }
}

# Barajar las posiciones de las columnas
#
proc barajar_columnas {} {
    global texto numcolumnas posicion0 posicion1 posicion2
    set posicion0 [list]
    set posicion1 [list]
    set posicion2 [list]
    for {set k 0} {$k<$numcolumnas} {incr k} {
        lappend posicion0 $k
        lappend posicion1 -1
        lappend posicion2 -1
    }
    shuffle posicion0
    destroy .texto.bajo
    crear_celdas_bajo
}

#============================================================
#          Desarrollo del juego
#============================================================

# Comprobar exito
#
proc exito {} {
    global numcolumnas posicion1
    set ok true
    for {set col 0} {$col < $numcolumnas && $ok} {incr col} {
        set ok [expr [lindex $posicion1 $col] == $col]
    }
    return $ok
}

# Iniciar el juego con un nuevo fragmento de texto
#
proc nuevo_texto {} {
    global texto fichero_texto longtexto
    set texto [read_fragment $fichero_texto $longtexto $longtexto]
    barajar_columnas
    reiniciar_juego
}

# Reiniciar el juego con el mismo fragmento de texto
#
proc reiniciar_juego {} {
    global numcolumnas posicion1 posicion2 seleccion_alto seleccion_bajo
    set posicion1 [list]
    set posicion2 [list]
    for {set k 0} {$k<$numcolumnas} {incr k} {
        lappend posicion1 -1
        lappend posicion2 -1
    }
    set seleccion_alto -1
    set seleccion_bajo -1
    actualizar_todo
    mostrar_exito false
}

# Seleccionar una columna del texto a reonstruir
#
proc seleccionar_alto {col {ok 0}} {
    global seleccion_bajo seleccion_alto posicion2
    set seleccion_alto_antes $seleccion_alto
    set seleccion_bajo_antes $seleccion_bajo
    set modo [modo_boton_alto $col]
    if {$modo == "desmarcar" && !$ok} {      # quitar seleccin
        set seleccion_alto -1
        set seleccion_bajo -1
    } else {                                 # poner seleccin
        set seleccion_alto $col
        set origen [lindex $posicion2 $col]
        if {$origen >= 0} {
            set seleccion_bajo $origen
        }
    }
    actualizar_cambio $seleccion_alto_antes $seleccion_alto $seleccion_bajo_antes $seleccion_bajo
}

# Retirar una columna recolocada por el jugador
#
proc bajar {col} {
    global posicion1 posicion2 seleccion_bajo
    set origen [lindex $posicion2 $col]
    if {$origen >= 0} {
        set posicion1 [lreplace $posicion1 $origen $origen -1]
        set posicion2 [lreplace $posicion2 $col $col -1]
        set seleccion_bajo_antes $seleccion_bajo
        set seleccion_bajo $origen
        actualizar_cambio $col -1 $seleccion_bajo_antes $origen
    }
}

# Colocar una columna en la posicin que indique el jugador
#
proc subir {col destino} {
    global posicion1 posicion2 seleccion_alto seleccion_bajo
    if {$destino >= 0} {
        set pos_antes [lindex $posicion1 $col]
        bajar $pos_antes    ;# retirar la columna a subir, si ya estaba colocada
        bajar $destino      ;# despejar destino, si es necesario
        # colocar la columna origen en el destino
        set posicion1 [lreplace $posicion1 $col $col $destino]
        set posicion2 [lreplace $posicion2 $destino $destino $col]
        set seleccion_alto $destino
        set seleccion_bajo_antes $seleccion_bajo
        set seleccion_bajo $col
        set ok [exito]
        if {$ok} {
            mostrar_exito true
            set seleccion_alto -1
            set seleccion_bajo -1
        }
        actualizar_cambio $pos_antes $destino $seleccion_bajo_antes $col
    }
}

proc modo_boton_alto {col} {
    global seleccion_alto
    if {$col == $seleccion_alto} {           # quitar seleccin
        return desmarcar
    } else {                                 # poner seleccin
        return marcar
    }
}

proc modo_boton_bajo {col} {
    global posicion1
    if {[lindex $posicion1 $col] >= 0} {     # columna ya subida
        set modo buscar
    } else {                                 # cambiar marca
        set modo subir
    }
    return $modo
}

proc operar_boton_alto {col} {
    seleccionar_alto $col
}

proc operar_boton_bajo {col} {
    global seleccion_alto posicion1
    switch [modo_boton_bajo $col] {
        subir  {subir $col $seleccion_alto}
        buscar {seleccionar_alto [lindex $posicion1 $col] 1}
    }
}

proc mover_izquierda {} {
    global seleccion_alto posicion1 posicion2 numcolumnas
    set col $seleccion_alto
    set origen [lindex $posicion2 $col]
    set destino [expr $col-1]
    if {$origen < 0 || $destino < 0 || [lindex $posicion2 $destino] >= 0} return
    set seleccion_alto $destino
    while {$origen >= 0} {
        lset posicion2 $destino $origen
        lset posicion1 $origen $destino
        lset posicion2 $col -1
        incr col
        incr destino
        if {$col >= $numcolumnas} {
            set origen -1
        } else {
            set origen [lindex $posicion2 $col]
        }
    }
    actualizar_todo
}

proc mover_derecha {} {
    global seleccion_alto posicion1 posicion2 numcolumnas
    set col $seleccion_alto
    set origen [lindex $posicion2 $col]
    set destino [expr $col+1]
    if {$origen < 0 || $destino >= $numcolumnas || [lindex $posicion2 $destino] >= 0} return
    set seleccion_alto $destino
    while {$origen >= 0} {
        lset posicion2 $destino $origen
        lset posicion1 $origen $destino
        lset posicion2 $col -1
        incr col -1
        incr destino -1
        if {$col < 0} {
            set origen -1
        } else {
            set origen [lindex $posicion2 $col]
        }
    }
    actualizar_todo
}

proc mover_abajo {} {
    global seleccion_alto
    bajar $seleccion_alto
}


#============================================================
#          Interfaz de usuario
#============================================================

#      Ttulo
# -----------------------------------------------------------
set titulo_normal "COLUMNAS MOVEDIZAS"
set titulo_exito "COLUMNAS MOVEDIZAS - TEXTO DESCIFRADO!"
set fuente_titulo [font create -size 14 -weight bold]
label .titulo -text $titulo_normal -font $fuente_titulo -padx 20 -pady 10
pack .titulo

proc mostrar_exito {exito} {
    global titulo_normal titulo_exito
    if {$exito} {
        .titulo configure -text $titulo_exito -foreground red
        set seleccion_alto -1
        set seleccion_bajo -1
    } else {
        .titulo configure -text $titulo_normal -foreground black
    }
}

#      Nombre del fichero de texto
# -----------------------------------------------------------
frame .texto ;# -padx 20
pack .texto -after .titulo
frame .texto.fichero
pack .texto.fichero -fill x -expand yes -pady 5 -padx 20
label .texto.fichero.nombre -text [file tail $fichero_texto]
pack .texto.fichero.nombre -side left

#      Columnas de la parte superior
# -----------------------------------------------------------
proc nombre_celda {fil col} {
    return "f$fil[]c$col"
}

proc letra_celda {fil col} {
    global texto numcolumnas
    if {$col>=0} {
        set off [expr $fil*$numcolumnas + $col]
        return [string index $texto $off]
    } else {
        return ""
    }
}

set marca_boton(marcar) "x"
set marca_boton(desmarcar) "o"
set marca_boton(subir) subir
set marca_boton(buscar) pregunta

set negrita [font create -weight bold]
proc crear_celdas_alto {} {
    global negrita numfilas numcolumnas
    frame .texto.alto
    pack .texto.alto -after .texto.fichero -pady 10 -padx 20
    for {set col 0} {$col < $numcolumnas} {incr col} {
        for {set fil 0} {$fil < $numfilas} {incr fil} {
            set celda [nombre_celda $fil $col]
            label .texto.alto.$celda -width 2 -relief solid -borderwidth 1 -font negrita  ;# -text $letra
            grid .texto.alto.$celda -column $col -row $fil
        }
        button .texto.alto.x$col -borderwidth 1 -text "x" -font negrita -command "operar_boton_alto $col"
        grid .texto.alto.x$col -column $col -row $numfilas
    }
}

#      Botones centrales
# -----------------------------------------------------------
proc crear_botones_centro {} {
    global negrita
    frame .texto.centro
    pack .texto.centro -after .texto.alto
    button .texto.centro.izquierda -font negrita -bitmap @izquierda.xbm -command "mover_izquierda" -height 28 -width 40
    button .texto.centro.bajar -font negrita -bitmap @bajar.xbm -command "mover_abajo" -height 28 -width 30
    button .texto.centro.derecha -font negrita -bitmap @derecha.xbm -command "mover_derecha" -height 28 -width 40
    pack .texto.centro.izquierda .texto.centro.bajar .texto.centro.derecha -side left -padx 2m
}

#      Columnas de la parte inferior
# -----------------------------------------------------------
proc crear_celdas_bajo {} {
    global negrita numfilas numcolumnas posicion0 marca_boton
    frame .texto.bajo
    pack .texto.bajo -after .texto.centro -pady 20 -padx 20
    for {set col 0} {$col < $numcolumnas} {incr col} {
        set col0 [lindex $posicion0 $col]
        button .texto.bajo.x$col0 -relief solid -borderwidth 1 -bitmap @$marca_boton([modo_boton_bajo $col]).xbm -command "operar_boton_bajo $col0" -height 28 -width 20
        grid .texto.bajo.x$col0 -column $col -row 0
        for {set fil 0} {$fil < $numfilas} {incr fil} {
            set celda [nombre_celda $fil $col0]
            set letra [letra_celda $fil $col0]
            label .texto.bajo.$celda -width 2 -text $letra -relief solid -borderwidth 1 -font negrita
            if {$letra == " "} {.texto.bajo.$celda configure -background black}
            grid .texto.bajo.$celda -column $col -row [expr $fil+1]
        }
    }
}

proc actualizar_columna_alto {col} {
    global numfilas numcolumnas posicion2 seleccion_alto marca_boton
    if {$col >= 0 && $col < $numcolumnas} {
        if {$col == $seleccion_alto} {
            set bg yellow
        } else {
            set bg white
        }
        set origen [lindex $posicion2 $col]
        for {set fil 0} {$fil < $numfilas} {incr fil} {
            set celda [nombre_celda $fil $col]
            set letra [letra_celda $fil $origen]
            .texto.alto.$celda configure -text $letra -background $bg
            if {$letra == " "} {.texto.alto.$celda configure -background black}
        }
        .texto.alto.x$col configure -text $marca_boton([modo_boton_alto $col])
    }
}

proc actualizar_columna_bajo {col} {
    global numfilas numcolumnas posicion1 seleccion_bajo marca_boton
    if {$col >= 0 && $col < $numcolumnas} {
        if {[lindex $posicion1 $col] >= 0} {
            set negro silver
        } else {
            set negro black
        }
        set fg $negro
        if {$col == $seleccion_bajo} {
            set fondo yellow
        } else {
            set fondo white
        }
        for {set fil 0} {$fil < $numfilas} {incr fil} {
            set celda [nombre_celda $fil $col]
            set letra [letra_celda $fil $col]
            if {$letra == " "} {
                set bg $negro
            } else {
                set bg $fondo
            }
            .texto.bajo.$celda configure -foreground $fg -background $bg
        }
        .texto.bajo.x$col configure -bitmap @$marca_boton([modo_boton_bajo $col]).xbm
    }
}

proc actualizar_botones {} {
    global numcolumnas marca_boton posicion1
    set exito true
    for {set col 0} {$col < $numcolumnas} {incr col} {
        .texto.bajo.x$col configure -text $marca_boton([modo_boton_bajo $col])
        if {[lindex $posicion1 $col] != $col} {set exito false}
    }
    return $exito
}

proc actualizar_cambio {alto1 alto2 bajo1 bajo2} {
    actualizar_columna_alto $alto1
    actualizar_columna_alto $alto2
    actualizar_columna_bajo $bajo1
    actualizar_columna_bajo $bajo2
}

proc actualizar_todo {} {
    global numcolumnas 
    for {set col 0} {$col < $numcolumnas} {incr col} {
        actualizar_columna_alto $col
        actualizar_columna_bajo $col
    }
}

#      Texto de ayuda
# -----------------------------------------------------------
proc ayuda {} {
    tk_messageBox -type ok -message {
Se trata de reconstruir un fragmento de texto en el que las 
columnas del texto se han desordenado. Para reconstruir el texto 
hay que ir colocando cada columna del texto desordenado de la 
parte inferior en la posicin correcta de la parte superior. 

Para poner una columna en la parte superior hay que seleccionar 
la columna en la parte inferior y pulsar el botn de subir (^) en 
la columna de la parte superior. O bien seleccionar una columna 
de la parte superior y pulsar el botn de subir (^) de la columna 
de la pare inferior. 

Para seleccionar o deseleccionar una columna hay que pulsar el 
botn de esa columna con el smbolo de marcar (x). Tambin se 
selecciona una columna ocupada de la parte superior pulsando el 
botn de localizacin (?) de la columna correspondiente de la 
parte inferior. 

Para retirar una columna ocupada de la parte superior hay que 
pulsar su botn de bajar (v) o bien el botn de bajar (v) en la 
columna correspondiente de la parte inferior. 

Las funciones del men superior son:

- Nuevo fichero: Permite elegir el fichero de texto del que
  se tomarn los fragmento de texto a reconstruir.

- Nuevo texto: Reinicia el juego con un nuevo fragmento de texto
  del fichero actual, con las columnas desordenadas de otra manera.
  
- Reiniciar: Reinicia el juego con el mismo fragmente de texto
  y desorden de las columnas.

- Ayuda: Presenta este mismo texto de ayuda.
  
- Salir: Termina el programa.
    }
}

#      Opciones de menu
# -----------------------------------------------------------
menu .menu
.menu add command -command nuevo_fichero -label "Nuevo fichero"
.menu add command -command nuevo_texto -label "Nuevo fragmento"
.menu add command -command reiniciar_juego -label "Reiniciar"
.menu add command -command ayuda -label "Ayuda"
.menu add command -command exit -label "Salir"
. configure -menu .menu


#      Iniciar el juego
# -----------------------------------------------------------

crear_celdas_alto
crear_botones_centro
nuevo_texto
