#============================================================
#          MINAS
#
#  Manuel Collado, http://mcollado.z15.es
#  Marzo, 2020 - Public domain
#
#============================================================
package require Tk
set rootdir [file normalize [file dirname [info script]]/..]
source $rootdir/lib/common.tcl

#============================================================
#          Modelo de datos
#============================================================

#      Estado del juego
# -----------------------------------------------------------
# -- Opciones
set numminas 30
set aspecto cuadrado
set numfilas 16
set numcolum 16
set nminas $numminas   ;# nuevo nmero de minas
set naspecto $aspecto  ;# nuevo aspecto
set nfilas $numfilas   ;# nuevo nmero de filas
set ncolum $numcolum   ;# nuevo nmero de columnas

# -- Campo minado
set minas ""           ;# campo minado      "..001X102X1.."   {minasvecinas(0-8) / haymina(X)}
set vista ""           ;# zona descubierta  "  ++++X....  "   {pisado(+) /marcado(X) / oculto(.)}
set filx -1            ;# mina pisada (fila)
set colx -1            ;# mina pisada (columna)

# -- Estado global del juego
set exito 0            ;# 0 = no terminado; 1 = xito; 2 = fallo

# Nmero de casillas vecinas de un tipo dado
#
proc vecinos {capa fil col marca} {
    set pos0 [indice $fil $col]
    set cuenta 0
    for {set f [expr $fil-1]} {$f <= [expr $fil+1]} {incr f} {
        for {set c [expr $col-1]} {$c <= [expr $col+1]} {incr c} {
            if {[dentro $f $c]} {
                set pos [indice $f $c]
                if {[string index $capa $pos] == $marca && $pos != $pos0} {
                    incr cuenta
                }
            }
        }
    }
    return $cuenta
}

# Comprobar ndices dentro de rango
#
proc dentro {fil col} {
    global numfilas numcolum
    return [expr $fil>=0 && $fil<$numfilas && $col>=0 && $col<$numcolum]
}

# Posicin global de una casilla en la coleccin lineal de casillas
#
proc indice {fil col} {
    global numcolum
    if {[dentro $fil $col]} {
        return [expr $fil*$numcolum + $col]
    } else {
        return -1
    }
}

# Mostrar una capa ****************************** PRUEBA ******************
#
proc mostrar {capa} {
    global numcolum
    set total [string length $capa]
    for {set k 0} {$k<$total} {incr k} {
        if {$k%$numcolum==0} {puts ""}
        puts -nonewline [string index $capa $k]
    }
    puts ""
}

#============================================================
#          Desarrollo del juego
#============================================================

# Iniciar el juego con una nueva distribucin de minas
#
proc iniciar_juego {} {
    global numfilas numcolum numminas minas vista filx colx exito
    
    # Repartir minas al azar e iniciar vista vaca
    set numceldas [expr $numfilas*$numcolum]
    set minas "."                 ;# primera casilla sin mina
    set vista "."
    set restoceldas [expr $numceldas-2]   ;# casillas intermedias
    set restominas $numminas
    for {set k 2} {$k < $numceldas} {incr k} {   # casillas intermedias
        if {$restoceldas*rand() < $restominas} {
            append minas "X"
            incr restominas -1
        } else {
            append minas "."
        }
        append vista "."
        incr restoceldas -1
    }
    append minas "."              ;# ltima casilla sin mina
    append vista "."

    # Precalcular nmero de minas vecinas
    set nuevo ""
    for {set fil 0} {$fil < $numfilas} {incr fil} {
        for {set col 0} {$col < $numcolum} {incr col} {
            set pos [indice $fil $col]
            if {[string index $minas $pos] == "X"} {
                append nuevo "X"
            } else {
                append nuevo [vecinos $minas $fil $col "X"]
            }
        }
    }
    set minas $nuevo
    set exito 0
    set filx -1
    set colx -1
}

# Destapar una casilla
#
proc pisar_casilla {fil col} {
    global numfilas numcolum minas vista filx colx exito
    set pos [indice $fil $col]
    set v [string index $vista $pos]
    if {$pos >= 0 && $v == "." && [vecinos $vista $fil $col "+"] > 0 || $pos == 0} {
        set m [string index $minas $pos]
        if {$m == "X"} {                  # ****** BOOM! ******
            set vista ""
            set filx $fil
            set colx $col
            destapar_todo
            set exito -1
            mostrar_exito -1
        } else {
            set vista [string replace $vista $pos $pos "+"]
            actualizar_celda $fil $col
            if {$m == 0} {
                propagar $fil $col
            }
            if {$fil==$numfilas-1 && $col==$numcolum-1} {
                set exito 1
                mostrar_exito 1
                actualizar_celda $fil $col
            }
        }
    }
}

# Destapar casillas vecinas sin minas vecinas
#
proc propagar {fil col} {
    for {set f [expr $fil-1]} {$f <= [expr $fil+1]} {incr f} {
        for {set c [expr $col-1]} {$c <= [expr $col+1]} {incr c} {
            if {$f != $fil || $c != $col} {
                pisar_casilla $f $c
           }
        }
    }
}

# Marcar/desmarcar una casilla como mina
#
proc marcar_casilla {fil col} {
    global vista exito
    if {$exito == 0} {
        set pos [indice $fil $col]
        set v [string index $vista $pos]
        if {$v == "."} {
            set vista [string replace $vista $pos $pos "X"]
        } elseif {$v == "X"} {
            set vista [string replace $vista $pos $pos "."]
        }
        actualizar_celda $fil $col
    }
}

# Destapar_todo, al final del juego
#
proc destapar_todo {} {
    global vista
    set vista ""
    actualizar_todo
}

proc actualizar_todo {} {
    global numfilas numcolum
    for {set fil 0} {$fil < $numfilas} {incr fil} {
        for {set col 0} {$col < $numcolum} {incr col} {
            actualizar_celda $fil $col
        }
    }
}

# Nuevo juego, quiz con nuevas opciones
#
proc nuevo_juego {numfil numcol nummin} {
    global numfilas numcolum numminas
    mostrar_exito 0
    set numminas $nummin
    if {$numfil != $numfilas || $numcol != $numcolum} {
        set numfilas $numfil
        set numcolum $numcol
        destroy .campo
        crear_campo
    }
    iniciar_juego
    actualizar_todo
    pisar_casilla 0 0
}

#============================================================
#          Interfaz de usuario
#============================================================

#      Ttulo
# -----------------------------------------------------------
set titulo_normal "CAMPO MINADO"
set titulo_exito "CAMPO MINADO - RESUELTO!"
set titulo_fallo "CAMPO MINADO - FALLO!"
set fuente_titulo [font create -size 14 -weight bold]
label .titulo -text $titulo_normal -font $fuente_titulo -padx 20 -pady 10
pack .titulo

proc mostrar_exito {exito} {
    global titulo_normal titulo_exito titulo_fallo
    if {$exito>0} {
        .titulo configure -text $titulo_exito -foreground green
    } elseif {$exito<0} {
        .titulo configure -text $titulo_fallo -foreground red
    } else {
        .titulo configure -text $titulo_normal -foreground black
    }
    activar_campo [expr $exito==0]
}

#      Campo minado
# -----------------------------------------------------------
set fondo_normal [. cget -background]
proc crear_campo {} {
    global numfilas numcolum
    set negrita [font create -weight bold]
    set lado_celda 15
    set alto_campo [expr $numfilas * $lado_celda + 4]
    set ancho_campo [expr $numcolum * $lado_celda + 4]
    frame .campo -relief solid -borderwidth 3 -width $ancho_campo[]p -height $alto_campo[]p
    pack .campo -after .titulo -pady 20 -padx 20
    for {set fil 0} {$fil < $numfilas} {incr fil} {
        for {set col 0} {$col < $numcolum} {incr col} {
            set celda "f$fil[]c$col"
            button .campo.$celda -borderwidth 3 -command "pisar_casilla $fil $col" -font $negrita -disabledforeground black
            set posy [expr $fil*$lado_celda]
            set posx [expr $col*$lado_celda]
            place .campo.$celda -y $posy[]p -x $posx[]p -width $lado_celda[]p -height $lado_celda[]p
            bind .campo.$celda <ButtonRelease-3> "marcar_casilla $fil $col"
        }
    }
    
}

set marca(.) " "
set marca(0) " "
set marca(1) "1"
set marca(2) "2"
set marca(3) "3"
set marca(4) "4"
set marca(5) "5"
set marca(6) "6"
set marca(7) "7"
set marca(8) "8"
set marca(X) "o'"
proc actualizar_celda {fil col} {
    global minas vista marca filx colx fondo_normal exito
    set fondo $fondo_normal
    set final [expr [string length $minas] - 1]
    set pos [indice $fil $col]
    set debajo [string index $minas $pos]
    if {$debajo==""} {set debajo 0}
    set encima [string index $vista $pos]
    if {$encima==""} {set encima "+"}
    set celda "f$fil[]c$col"
    if {$encima == "+"} {
        set relieve sunken
        set letra $marca($debajo)
        if {$pos==0 || $pos==$final && $exito>0} {set fondo green}
    } else {
        set relieve raised
        set letra $marca($encima)
    }
    if {$letra==$marca(X)} {
        .campo.$celda configure -relief $relieve -bitmap @bomba.xbm
    } else {
        .campo.$celda configure -relief $relieve -text $letra -bitmap {}
    }
    if {$fil==$filx && $col==$colx} {
        .campo.$celda configure -background red
    } else {
        .campo.$celda configure -background $fondo
    }
}

proc activar_campo {ok} {
    if {$ok} {
        set estado normal
    } else {
        set estado disabled
    }
    set botones [info commands .campo.*]
    foreach boton $botones {
        $boton configure -state $estado
    }
}

#      Texto de ayuda
# -----------------------------------------------------------
proc ayuda {} {
    tk_messageBox -type ok -message {
Se trata de ir pisando las casillas que forman un campo minado, 
con el objetivo de permitir el paso desde la casilla superior 
izquierda hasta la casilla inferior derecha, sin pisar ninguna 
mina. Las casillas pisadas quedan al descubierto mostrando el 
nmero de minas que hay en las casillas vecinas (no descubiertas 
todava). 

El juego empieza con todas las casillas tapadas excepto la 
primera, y termina con xito si se llega a pisar la ltima 
casilla, o con fallo si se pisa una casilla con una mina. En cada 
momento slo se admite pisar y destapar una casilla contigua a 
otra ya descubierta. 

Para pisar y destapar una casilla hay que pulsar con el botn 
derecho sobre ella. Pulsando con el botn derecho se pone o quita 
una marca para indicar que esa casilla debe contener una mina 
oculta. Esa marca asegura adems que la casilla no puede ser 
pisada accidentalmente. 

Las funciones del men superior son:

- Nuevo juego: Reinicia el juego con una nueva distribucin de minas.
  
- Opciones: Permite establecer el nivel de dificultad, fijando el 
  tamao del campo minado y el nmero de minas. 

- Ayuda: Presenta este mismo texto de ayuda.
  
- Salir: Termina el programa.
    }
}

#      Opciones de menu
# -----------------------------------------------------------
menu .menu
.menu add command -command {nuevo_juego $numfilas $numcolum $numminas} -label "Nuevo juego"
.menu add command -command opciones -label "Opciones"
.menu add command -command ayuda -label "Ayuda"
.menu add command -command exit -label "Salir"
. configure -menu .menu

#      Dilogo de opciones
# -----------------------------------------------------------
proc opciones {} {
    global fuente_titulo nminas naspecto nfilas ncolum numminas aspecto numfilas numcolum
    
    # Crear la ventana de dilogo
    toplevel .opciones
    label .opciones.titulo -text "Nivel del juego" -font $fuente_titulo -padx 20 -pady 10
    pack .opciones.titulo
    
    frame .opciones.entrada
    pack .opciones.entrada -padx 20
    
    label .opciones.entrada.labelminas -text "Numero de minas:"
    entry .opciones.entrada.entryminas -width 10 -textvariable nminas
    grid .opciones.entrada.labelminas .opciones.entrada.entryminas -row 0
    
    set cuadrado {set ncolum $nfilas; .opciones.entrada.entrycolum configure -state disabled}
    set rectangulo {set ncolumn $numcolum; .opciones.entrada.entrycolum configure -state normal}
    radiobutton .opciones.entrada.cuadro -variable naspecto -text "Cuadrado" -value "cuadrado" -command $cuadrado
    radiobutton .opciones.entrada.rectangulo -variable naspecto -text "Rectangulo" -value "rectangulo" -command $rectangulo
    grid .opciones.entrada.cuadro .opciones.entrada.rectangulo -row 1
    
    label .opciones.entrada.labelfilas -text "Numero de filas:"
    entry .opciones.entrada.entryfilas -width 10 -textvariable nfilas
    grid .opciones.entrada.labelfilas .opciones.entrada.entryfilas -row 2
    
    label .opciones.entrada.labelcolum -text "Numero de columnas:"
    entry .opciones.entrada.entrycolum -width 10 -textvariable ncolum
    grid .opciones.entrada.labelcolum .opciones.entrada.entrycolum -row 3
    if {$naspecto=="cuadrado"} {eval $cuadrado}
    
    frame .opciones.botones
    pack .opciones.botones -pady 10
    button .opciones.botones.aceptar -text "Aceptar" -command "aceptar_opciones"
    button .opciones.botones.cancelar -text "Cancelar" -command "cancelar_opciones"
    set mostrar {tk_messageBox -type ok -parent .opciones -message "$nminas $naspecto $nfilas $ncolum\n$numminas $aspecto $numfilas $numcolum"}
    button .opciones.botones.mostrar -text "Mostrar" -command $mostrar
    pack .opciones.botones.aceptar .opciones.botones.cancelar .opciones.botones.mostrar -side left
    
    center_window .opciones
    make_modal .opciones cancelar_opciones
}

proc aceptar_opciones {} {
    global nminas naspecto nfilas ncolum numminas aspecto numfilas numcolum

    set error ""
    set maxminas [expr $nfilas*$ncolum/4]
    set maxfilas 32
    set maxcolum 60
    if {$nfilas<=0 || $nfilas>$maxfilas} {
        set error "Numero de filas: $nfilas - debe ser (1-$maxfilas)"
    } elseif {$naspecto=="cuadrado"} {
        set ncolum $nfilas
    } elseif {$ncolum<=0 || $ncolum>$maxcolum} {
        set error "Numero de columnas: $ncolum - debe ser (1-$maxcolum)"
    }
    if {$error=="" && ($nminas<=0 || $nminas>$maxminas)} {
        set error "Numero de minas: $nminas - debe ser (1-$maxminas)"
    }
    if {$error==""} {
        destroy .opciones
        set aspecto $naspecto
        nuevo_juego $nfilas $ncolum $nminas
    } else {
        tk_messageBox -type ok -parent .opciones -icon warning -title "OPCION NO VALIDA" -message $error ;# -parent .opciones
    }
}

proc cancelar_opciones {} {
    global numminas aspecto numfilas numcolum nminas naspecto nfilas ncolum

    destroy .opciones
    set nminas $numminas
    set naspecto $aspecto
    set nfilas $numfilas
    set ncolum $numcolum
}

#      Iniciar el juego
# -----------------------------------------------------------
crear_campo
iniciar_juego
actualizar_todo
pisar_casilla 0 0
