#============================================================
#          TEXTO OCULTO
#
#  Manuel Collado, http://mcollado.z15.es
#  Marzo, 2020 - Public domain
#
#============================================================
package require Tk
set rootdir [file normalize [file dirname [info script]]/..]
source $rootdir/lib/common.tcl

#============================================================
#          Modelo de datos
#============================================================

#      Clave y texto cifrado
# -----------------------------------------------------------
set letras "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
set numletras [string length $letras]
set letra [list]
set clave [list]
set letra0 [list]
set seleccion 0
set numfilas 11
set numcolumnas 22
set numchars [expr $numfilas * $numcolumnas]
set texto ""
set cifra [list]

proc crear_clave {} {
    global letras numletras letra clave letra0 pos_letra pos_letra pos_letra0 pos_clave
    set letra [list]
    set clave [list]
    set letra0 [list]
    for {set k 0} {$k < $numletras} {incr k} {
        set c [string index $letras $k]
        lappend letra $c
        lappend clave [expr $k+1]
        lappend letra0 ""
        set pos_letra($c) $k
        set pos_letra0($c) -1
    }
    shuffle clave
    for {set k 0} {$k < $numletras} {incr k} {
        set pos_clave([lindex $clave $k]) $k
    }
}

#      Desarrolo del juego
# -----------------------------------------------------------
proc reiniciar_juego {} {
    global numletras letra0
    set letra0 [list]
    for {set k 0} {$k < $numletras} {incr k} {
        lappend letra0 ""
    }
    mostrar_exito false
    seleccionar 0
    reiniciar_celdas_texto
    reiniciar_botones
}

proc seleccionar {codigo} {
    global seleccion
    if {$codigo != $seleccion && $seleccion != 0} {
        marcar_celdas $seleccion false
    }
    set seleccion $codigo
    if {$codigo > 0} {
        marcar_celdas $seleccion true
    }
}

proc asociar { letra } {
    global letra0 pos_letra0 clave pos_clave seleccion
    set pos_let0 $pos_letra0($letra)
    if {$pos_let0 >= 0} {
        set letra0 [lreplace $letra0 $pos_let0 $pos_let0 ""]
        set pos_letra0($letra) -1
        set codigo0 [lindex $clave $pos_let0]
        actualizar_celdas $codigo0
        subir_boton $letra
    }
    if {$seleccion > 0} {
        set pos_cod $pos_clave($seleccion)
        set let0 [lindex $letra0 $pos_cod]
        if {$let0 != ""} {
            set pos_letra0($let0) -1
            subir_boton $let0
        }
        set letra0 [lreplace $letra0 $pos_cod $pos_cod $letra]
        set pos_letra0($letra) $pos_cod
        actualizar_celdas $seleccion
        bajar_boton $letra
    }
}


#      Texto cifrado
# -----------------------------------------------------------

# Elegir fichero con textos
#
set fichero_texto "$rootdir/textos/texto.txt"
proc nuevo_fichero {} {
    global fichero_texto
    set fichero [new_file $fichero_texto]
    if {$fichero != ""} {
        set fichero_texto $fichero
       .texto.fichero.nombre configure -text [file tail $fichero_texto]
        nuevo_texto
    }
}

# Leer y cifrar fragmento de texto
#
proc leer_texto {fichero} {
    global numchars texto cifra clave pos_letra
    set cifra [list]
    set texto [read_fragment $fichero $numchars $numchars]
    for {set k 0} {$k < $numchars} {incr k} {
        set c [string index $texto $k]
        if {[info exists pos_letra($c)]} {
            set codigo [lindex $clave $pos_letra($c)]
            lappend cifra $codigo
        } else {
            lappend cifra 0
        }
    }
}

proc nuevo_texto {} {
    global fichero_texto
    crear_clave
    leer_texto $fichero_texto
    rehacer_celdas_texto
    reiniciar_juego
}


#============================================================
#          Interfaz de usuario
#============================================================

#      Ttulo
# -----------------------------------------------------------
set titulo_normal "TEXTO OCULTO"
set titulo_exito "TEXTO OCULTO - TEXTO DESCIFRADO!"
set fuente_titulo [font create -size 14 -weight bold]
label .titulo -text $titulo_normal -font $fuente_titulo -padx 20 -pady 10
pack .titulo

proc mostrar_exito {exito} {
    global titulo_normal titulo_exito
    if {$exito} {
        .titulo configure -text $titulo_exito -foreground red
        seleccionar 0
    } else {
        .titulo configure -text $titulo_normal -foreground black
    }
}

#      Fichero de texto y celdas con el texto cifrado
# -----------------------------------------------------------
frame .texto ;# -padx 20
pack .texto -after .titulo
frame .texto.fichero
pack .texto.fichero -fill x -expand yes -pady 5 -padx 20
label .texto.fichero.nombre -text [file tail $fichero_texto]
pack .texto.fichero.nombre -side left

proc clave_celda { k } {
    global cifra
    set codigo [lindex $cifra $k]
    if {$codigo == 0} {
        set codigo ""
    }
    return $codigo
}

proc letra_celda { k } {
    global texto cifra pos_letra letra0
    set codigo [lindex $cifra $k]
    set caracter [string index $texto $k]
    if {$codigo != 0} {
        set pos_caracter $pos_letra($caracter)
        set caracter0 [lindex $letra0 $pos_caracter]
        if {$caracter0 != ""} {
            set caracter $caracter0
        } else {
            set caracter " "
        }
    }
    return $caracter
}

set negrita [font create -weight bold]
proc crear_celdas_texto {} {
    global numchars texto cifra negrita numcolumnas
    frame .texto.celdas
    pack .texto.celdas -after .texto.fichero -pady "10"
    for {set k 0} {$k < $numchars} {incr k} {
        frame .texto.celdas.x$k -width 2 -borderwidth 1 -relief solid
        grid .texto.celdas.x$k -column [expr $k%$numcolumnas] -row [expr $k/$numcolumnas]
        button .texto.celdas.x$k.clave -width 2 -text [clave_celda $k] -command "seleccionar [lindex $cifra $k]" -relief solid -borderwidth 0 -pady 0
        button .texto.celdas.x$k.letra -width 2 -text [letra_celda $k] -command "seleccionar [lindex $cifra $k]" -relief solid -borderwidth 0 -pady 0 -font negrita
        pack .texto.celdas.x$k.clave .texto.celdas.x$k.letra -padx 0 -pady 0 -fill x -expand yes
        if {[string index $texto $k] == " "} {
            .texto.celdas.x$k.clave configure -background black
            .texto.celdas.x$k.letra configure -background black
        }
    }
}

proc rehacer_celdas_texto {} {
    destroy .texto.celdas
    crear_celdas_texto
}

proc actualizar_celda {k} {
    .texto.celdas.x$k.clave configure -text [clave_celda $k]
    .texto.celdas.x$k.letra configure -text [letra_celda $k]
}

proc actualizar_celdas {codigo} {
    global numchars cifra texto letra0 pos_clave
    set exito true
    for {set k 0} {$k < $numchars} {incr k} {
        set cifra_k [lindex $cifra $k]
        if {$codigo == $cifra_k} {
            actualizar_celda $k
        }
        if {$cifra_k > 0} {
            set pos_cifra $pos_clave($cifra_k)
            set letra0_k [lindex $letra0 $pos_cifra]
            set letra_k [string index $texto $k]
            if {$letra0_k != $letra_k} {
                set exito false
            }
        }
    }
    if {$exito} {
        mostrar_exito $exito
    }
}

proc reiniciar_celdas_texto {} {
    global numchars texto
    for {set k 0} {$k < $numchars} {incr k} {
        actualizar_celda $k
    }
}

set fondo_normal [. cget -background]
proc marcar_celdas {codigo ok} {
    global numchars cifra fondo_normal
    if {$ok} {
        set fondo yellow
    } else {
        set fondo $fondo_normal
    }
    for {set k 0} {$k < $numchars} {incr k} {
        if {[lindex $cifra $k] == $codigo} {
            .texto.celdas.x$k.clave configure -background $fondo
            .texto.celdas.x$k.letra configure -background $fondo
        }
    }
}

#      Botones con letras
# -----------------------------------------------------------
proc crear_botones {} {
    global letra numletras
    frame .botones ;# -padx 20 -pady 20
    pack .botones -after .texto -padx 20 -pady 20
    for {set k 0} {$k < $numletras} {incr k} {
        set c [lindex $letra $k]
        button .botones.x$c -text $c -command "asociar $c" -font negrita
        grid .botones.x$c -row 0 -column $k
    }
}

proc subir_boton { c } {
    .botones.x$c configure -relief raised
}

proc bajar_boton { c } {
    .botones.x$c configure -relief sunken
}

proc reiniciar_botones {} {
    global letra numletras
    for {set k 0} {$k < $numletras} {incr k} {
        set c [lindex $letra $k]
        subir_boton $c
    }
}

#      Texto de ayuda
# -----------------------------------------------------------
proc ayuda {} {
    tk_messageBox -type ok -message {
Se trata de descifrar un fragmento de texto en el que cada letra
se ha cifrado mediante un cdigo numrico. Para descifrar el texto
hay que ir probando a asociar cada cdigo numrico con la letra
que se considere que corresponde.

Para asociar una letra a un cdigo hay que seleccionar una casilla
del texto a descifrar y pulsar el botn inferior con la letra
que se desea asociar.

Las asociaciones de letras a cdigos pueden cambiarse cuanto sea
necesario. Para quitar una asociacin sin sustituirla por otra
basta seleccionar una casilla con espacio en blanco y pulsar el
botn de la letra a desasociar.

Las funciones del men superior son:

- Fichero de texto: Permite elegir el fichero de texto del que
  se tomarn los fragmento de texto a descifrar

- Nuevo texto: Reinicia el juego con un nuevo fragmento de texto
  del fichero actual, cifrado con una clave nueva.
  
- Reiniciar: Reinicia el juego con el mismo fragmente de texto
  y clave.

- Ayuda: Presenta este mismo texto de ayuda.
  
- Salir: Termina el programa.
    }
}

#      Opciones de menu
# -----------------------------------------------------------
menu .menu
.menu add command -command nuevo_fichero -label "Nuevo fichero"
.menu add command -command nuevo_texto -label "Nuevo texto"
.menu add command -command reiniciar_juego -label "Reiniciar"
.menu add command -command ayuda -label "Ayuda"
.menu add command -command exit -label "Salir"
. configure -menu .menu


#      Iniciar el juego
# -----------------------------------------------------------

crear_clave
leer_texto $fichero_texto
crear_celdas_texto
crear_botones
